package org.apache.bcel.generic;

import java.util.Arrays;

public final class SWITCH implements CompoundInstruction
{
    private int[] match;
    private InstructionHandle[] targets;
    private final Select instruction;
    private final int matchLength;
    
    public SWITCH(final int[] match, final InstructionHandle[] targets, final InstructionHandle target) {
        this(match, targets, target, 1);
    }
    
    public SWITCH(final int[] match, final InstructionHandle[] targets, final InstructionHandle target, final int maxGap) {
        this.match = match.clone();
        this.targets = targets.clone();
        final int length = match.length;
        this.matchLength = length;
        if (length < 2) {
            this.instruction = new TABLESWITCH(match, targets, target);
        }
        else {
            this.sort(0, this.matchLength - 1);
            if (this.matchIsOrdered(maxGap)) {
                this.fillup(maxGap, target);
                this.instruction = new TABLESWITCH(this.match, this.targets, target);
            }
            else {
                this.instruction = new LOOKUPSWITCH(this.match, this.targets, target);
            }
        }
    }
    
    private void fillup(final int maxGap, final InstructionHandle target) {
        final int maxSize = this.matchLength + this.matchLength * maxGap;
        final int[] mVec = new int[maxSize];
        final InstructionHandle[] tVec = new InstructionHandle[maxSize];
        int count = 1;
        mVec[0] = this.match[0];
        tVec[0] = this.targets[0];
        for (int i = 1; i < this.matchLength; ++i) {
            final int prev = this.match[i - 1];
            for (int gap = this.match[i] - prev, j = 1; j < gap; ++j) {
                mVec[count] = prev + j;
                tVec[count] = target;
                ++count;
            }
            mVec[count] = this.match[i];
            tVec[count] = this.targets[i];
            ++count;
        }
        this.match = Arrays.copyOf(mVec, count);
        this.targets = Arrays.copyOf(tVec, count);
    }
    
    public Instruction getInstruction() {
        return this.instruction;
    }
    
    @Override
    public InstructionList getInstructionList() {
        return new InstructionList(this.instruction);
    }
    
    private boolean matchIsOrdered(final int maxGap) {
        for (int i = 1; i < this.matchLength; ++i) {
            if (this.match[i] - this.match[i - 1] > maxGap) {
                return false;
            }
        }
        return true;
    }
    
    private void sort(final int l, final int r) {
        int i = l;
        int j = r;
        final int m = this.match[l + r >>> 1];
        while (true) {
            if (this.match[i] < m) {
                ++i;
            }
            else {
                while (m < this.match[j]) {
                    --j;
                }
                if (i <= j) {
                    final int h = this.match[i];
                    this.match[i] = this.match[j];
                    this.match[j] = h;
                    final InstructionHandle h2 = this.targets[i];
                    this.targets[i] = this.targets[j];
                    this.targets[j] = h2;
                    ++i;
                    --j;
                }
                if (i > j) {
                    break;
                }
                continue;
            }
        }
        if (l < j) {
            this.sort(l, j);
        }
        if (i < r) {
            this.sort(i, r);
        }
    }
}
